;; open Assert

(* In this lab you will be going over some basic debugging tools and practices.
 * In OCaml, we have three basic such tools:
 *   1) Manual Analysis
 *   2) Testing
 *   3) Using Print Statements
 *
 * The following lesson will be approximately split across these three topics.
 * Look over the example problems provided for each section and attempt the
 * practice problems below them. *)

(**-----------------------------------------------------------
 ** Manual Analysis
 **---------------------------------------------------------**)

(*----------------------------
 * example 1
 *--------------------------*)

(* Manual analysis is the process of breaking code down line-by-line with the
 * intention of fully understanding its functionality. Observe the following
 * example. *)

let rec sum (l: int list) : int =
  begin match l with
  | [] -> 0
  | x :: xs -> x + sum xs
  end

(* We break this function down as follows, given some selected starting
 * parameter [5;4;3;2;1]:
 *
 *    sum [5;4;3;2;1]
 *  -> 5 + sum [4;3;2;1] (because h = 5, tl = [4;3;2;1])
 *  -> 5 + 4 + sum [3;2;1] (h = 4, tl = [3;2;1])
 *  -> 5 + 4 + 3 + sum [2;1] (h = 3, tl = [2;1])
 *  -> 5 + 4 + 3 + 2 + sum [1] (h = 2, tl = [1])
 *  -> 5 + 4 + 3 + 2 + 1 + sum [] (h = 1, tl = [])
 *  -> 5 + 4 + 3 + 2 + 1 + 0
 *  -> 15
 *
 * Pretty simple, right? The next problem is for you to complete. *)

(*----------------------------
 * problem 1
 *--------------------------*)

(* Break this function down yourself to figure out what it is that it does. You
 * may write your answer on paper or comment it under the function. Note: the
 * `@` operator is built-in function that appends two lists together.*)

let rec your_turn (l: int list) : int list =
  begin match l with
  | [] -> []
  | _ :: xs -> your_turn xs @ [sum l]
  end

(* TO COMPLETE: your_turn [1; 2; 3] *)

(*----------------------------
 * problem 2
 *--------------------------*)

(* The following function determines whether or not a given integer list
 * contains alternating even/odd elements. Unfortunately, it isn't quite right.
 * Can you find the bug? Use manual analysis.
 *
 * Note: it is up to you to select an input string to analyze with. If you are
 * having trouble coming up with such an input that fails, skip ahead to the
 * testing section and come back later! *)

(* Helper: returns true if an integer x is even. *)
let even (x: int) : bool =
  x mod 2 = 0

let rec alternating (l: int list) : bool =
  begin match l with
  | [] -> true
  | [_] -> true
  | x1 :: x2 :: xs -> even x1 && not (even x2) && alternating (x2 :: xs)
  end

(**-----------------------------------------------------------
 ** Testing
 **---------------------------------------------------------**)

(* Sometimes code can get a little too long and/or complex to analyze by hand.
 * In these cases, testing is a great way to hone in on any potential bugs.
 * That's right, testing can do more than just discover bugs - it can also aid
 * you in addressing them!
 *
 * Okay, but how? Well, first, let's be sure you properly understand what it
 * means to test code. When you write a test, you are running your code with a
 * particular input value(s) and monitoring the result, ensuring that it
 * conforms with what you expect to see for that input. The goal is to vary
 * this input enough that you can be confident no possible input will result in
 * an error. Yet, for many functions there could be an infinite number of
 * possible inputs. Clearly you can't test all of them. To circumvent this
 * problem, notice that certain sets of input all cause your code to execute
 * in roughly the same way. Thus it is only necessary to test one such input
 * from each of these sets.
 *
 * As an example, let's take a look at the sum function again. What if we chose
 * a different list length with which to break the code down?
 *
 *   sum [3;2;1]
 *   -> 3 + sum [2;1]
 *   -> 3 + 2 + sum [1]
 *   -> 3 + 2 + 1 + sum []
 *   -> 3 + 2 + 1 + 0
 *   -> 6
 *
 * Because of recursion, the exact call, sum [3;2;1], was already incorporated
 * by the previous test of length 5, sum [5;4;3;2;1]. Thus the test is entirely
 * redundant! In fact, changing the length of test input, apart from the empty
 * case, does not make for interesting testing as they execute the code
 * identically.
 *
 * Similarly, if we try altering the integer values contained within the input
 * list,
 *
 *  sum [24;61;15]
 *  -> 24 + sum [61;15]
 *  -> 24 + 61 + sum [15]
 *  -> 24 + 61 + 15 + sum []
 *  -> 24 + 61 + 15 + 0
 *
 * We again see that the manner in which the code executes is identical. The
 * only difference is the values applied to the addition operator; however,
 * we can be reasonably confident that this operator works correctly (and it's
 * not our job to test it). Thus, this too is not an interesting test case.*)

(*----------------------------
 * problem 3
 *--------------------------*)

(* One way that a function can execute differently is through branching. This is
 * typically as a result of pattern matching and if/then/else statements. Some
 * input will cause your code to choose one branch over another. Thus one metric
 * for exhaustive testing is line coverage. In the following problem,
 * exhaustively test the function "branch_galore". Don't leave out any branches!
 *)

let foo (x: int) : string =
  if x < 0 then "As are"
  else "these"

let branch_galore (l: int list) : string =
  let len = List.length l in
  if len > 5 then "These branches"
  else if len > 4 then "are"
  else
    begin match l with
    | [] -> "mutually"
    | [_] -> "exclusive"
    | x1 :: x2 :: xs -> if (x1 - x2) < 0 then foo x2 else foo x1
    end

(*----------------------------
 * problem 4
 *--------------------------*)

(* Oftentimes, however, line coverage is not sufficient for exhaustive testing.
 * Anything from "strange" input values to input size could affect the way an
 * OCaml function executes. Attempt to find the bug in the following code by
 * exhaustively testing it. *)

(* This function returns true only if its input is
 * a perfect square--or does it?? *)
let square_check (x: int) : bool =
  let square = x * x in
  let original = square / x in
  original = x

(*----------------------------
 * problems 5,6
 *--------------------------*)

(* But what if you do discover a bug, a failing test? Well, if your tests are
 * written with specific behavior of your code in mind, then much of your work
 * may already be done! Good testing should reveal the logical nature of an
 * error. Provided your code is readable, making the step from logical to
 * practical understanding should be a simple matter of looking up the lines
 * that take care of said logic.
 *
 * The following two functions fail to execute properly. First, review the test
 * cases and their results. Based off of this information, determine the logical
 * nature of each bug. Then, use this knowledge to discover the actual bug in
 * the code. You may add additional test cases if you would like. *)

let rec factorial (x: int) : int =
  if x = 0 then 1 else x * factorial (x - 1)

let test () : bool =
  factorial 0 = 1
;; run_test "factorial: 0" test

let test () : bool =
  factorial 4 = 24
;; run_test "factorial: positive" test

(* Test error: Stack Overflow *)
let test () : bool =
  factorial (-2) = 1
;; run_test "factorial: negative" test

(* Note: if you feel lost, try manually analyzing the code with the failing
 * test parameters as input *)

let rec contains_multiple (l: int list) (x: int) : bool =
  begin match l with
  | [] -> false
  | y :: ys ->
      if y = x then contains_multiple ys x
      else contains_multiple ys x
  end

let test () : bool =
  contains_multiple [] 2 = false
;; run_test "contains_multiple: empty list" test

let test () : bool =
  contains_multiple [1; 3; 4] 2 = false
;; run_test "contains_multiple: does not contain" test

let test () : bool =
  contains_multiple [1; 2; 3; 4] 2 = false
;; run_test "contains_multiple: one" test

(* Test failed *)
let test () : bool =
  contains_multiple [1; 2; 2; 3; 4] 2 = true
;; run_test "contains_multiple: two" test

(* Test failed *)
let test () : bool =
  contains_multiple [2; 1; 2; 2; 3; 4; 2] 2 = true
;; run_test "contains_multiple: many" test

(**-----------------------------------------------------------
 ** Using Print Statements
 **---------------------------------------------------------**)

(* The third tool at your disposal is using print statements.
 * There are three basic print commands built into OCaml:
 *
 *  ;; print_endline "string"
 *  ;; print_string "string\n"
 *  ;; print_int 5
 *
 * Each of these will print text to the console. We will discuss OCaml commands
 * and printing from within a function later in the semester, but in the mean
 * time you may think of print statements as something you may run like a test,
 * outside of a function. Sometimes it is useful to see what a test case result
 * actually is (OCaml will only tell you what tests failed, not what results
 * those function calls gave). *)

(*----------------------------
 * problems 7,8
 *--------------------------*)

(* Practice debugging the functions that follow by adding print statements after
 * each test case, and then using all the tools now at your disposal to resolve
 * the bugs *)

(* A list printing function for your own use *)
let rec list_print (l: int list) : unit =
  begin match l with
  | [] -> print_endline "[]"
  | x :: xs -> print_int x; print_string " :: "; list_print xs
  end

let fractions (l: int list) : int =
  let rec loop (l2: int list) : int =
    begin match l2 with
    | [] -> 0
    | x :: xs -> (1 / x) + loop xs
    end in
  let fraction = loop l in
  factorial fraction

let test () : bool =
  fractions [] = 1
;; run_test "fractions: empty list" test

(* Example print statement. Complete the rest on your own *)
;; print_int (fractions [])

let test () : bool =
  fractions [1; 1; 1] = 6
;; run_test "fractions: multiple integer elements" test

let test () : bool =
  fractions [4; 2; 2] = 1
;; run_test "fractions: varied values" test

let test () : bool =
  fractions [4; 2; 2; 3; 3; 3] = 2
;; run_test "fractions: varied values2" test

(* The purpose of this function is, given an int list, ouput the difference
 * between adjacent values. For example: [1;2;3;5;4] -> [1;1;2;-1;0] *)

let rec derivative (l: int list) : int list =
  begin match l with
  | [] -> []
  | [_] -> [0]
  | x1 :: x2 :: xs -> x2 - x1 :: derivative xs
  end

let test () : bool =
  derivative [] = []
;; run_test "derivative: empty" test

let test () : bool =
  derivative [1] = [0]
;; run_test "derivative: singleton" test

let test () : bool =
  derivative [1; 2; 3; 5; 4] = [1; 1; 2; -1; 0]
;; run_test "derivative: full list" test

